//
//------------------------------------------------------------------------------
// Name    : rtttl.bas
// Author  : Jerry Messina
// Date    : 7/27/2021
// Version : 1.00
// Notes   : RingTone Text Transfer Language (RTTTL) library
//------------------------------------------------------------------------------
//
module RTTTL

//  
// The Nokia RingTone Text Transfer Language (RTTTL) is a method for
// specifying musical sequences as a string of text.
// It consists of a header comprised of the name and optional default 
// duration, scale, and tempo settings, followed by a series of note descriptions.
// It covers an abbreviate scale of four octaves from C4 through B7
//
// Format Specification:
//    <ringing-tones-text-transfer-language> :=
//       <name> <sep> [<defaults>] <sep> <note-command>+
//    
//    <name> := <char>+ ; maximum name length 10 characters**
//    
//    <sep> := ":"
//    
//    <defaults> := 
//       <def-note-duration> |
//       <def-note-scale> |
//       <def-beats> 
//    
//    <def-note-duration> := "d=" <duration>
//    
//    <def-note-scale> := "o=" <scale(octave)> 
//    
//    <def-beats> := "b=" <beats-per-minute>
//    
//    <beats-per-minute> := 25,28,...,900
//    
//    ; if not specified, the defaults are
//       ; 4 = duration
//       ; 6 = scale 
//       ; 63 = beats-per-minute
//    
//    <note-command> :=
//       [<duration>] <note> [<scale>] [<special-duration>] <delimiter>
//    
//    <duration> := "1" | "2" | "4" | "8" | "16" | "32" 
//       ; duration is divider of full note duration, eg. 4 represents a quarter note
//    
//    <note> := "C" | "C#" | "D" | "D#" | "E" | "F" | "F#" | "G" | "G#" | "A" | "A#" | "B" | "P" 
//       ; "P" is pause (or rest)
//    
//    <scale> :="4" | "5" | "6" | "7"
//       ; note that octave 4: A=440Hz, 5: A=880Hz, 6: A=1760Hz, 7: A=3520Hz
//    
//    <special-duration> :=
//       "."        ; dotted note (increase duration by 1.5)**
//    
//    <delimiter> := ","
//
// **SF implementation extensions
//  - allow <name> to be up to the SF string default of 23 chars
//    whitespace is not ignored in the RTTTL name section
//  - <special-duration> is supposed to appear AFTER the duration-note-scale (ie '8a#6.'),
//    but many examples have it BEFORE the scale (ie '8a#.6'), so both are allowed
//  - allow European scale "H" to represent USA scale 12th note "B"
//  - allow use of '-' char to denote 'flat' ie A# = B-
//  - accepts upper or lower-case
//
// RTTTL example:
// "HappyBday:d=4,o=5,b=125:8d.,16d,e,d,g,2f#,8p,8d.,16d,e,d,a,2g,8p,8d.,16d,d6,b,g,f#,2e,8p,8c6.,16c6,b,g,a,2g"
//
// usage:
//  declare a const string containing the RTTTL text and call Play with the string address
//
//  public const HappyBday = "HappyBday:d=4,o=5,b=125:8d.,16d,e,d,g,2f#,8p,8d.,16d,e,d,a,2g,8p,8d.,16d,d6,b,g,f#,2e,8p,8c6.,16c6,b,g,a,2g"
//  rtttl.Play(@HappyBday)
//
// for more info see-
//  https://www.mobilefish.com/tutorials/rtttl/rtttl_quickguide_specification.html
//  https://panuworld.net/nuukiaworld/download/nokix/rtttl.htm
//  https://en.wikipedia.org/wiki/Ring_Tone_Text_Transfer_Language
//

// use tone library to generate notes
include "tone.bas"

const 
    NOTES_PER_SCALE = 12,
    DEFAULT_SCALE = 4,
    OCTAVE_OFFSET = 0

// RTTTL notes (4 scales, from C4 to B7)
const notes() as word = 
( 
    0,    // pause
    NOTE_C4, NOTE_CS4, NOTE_D4, NOTE_DS4, NOTE_E4, NOTE_F4, NOTE_FS4, NOTE_G4, NOTE_GS4, NOTE_A4, NOTE_AS4, NOTE_B4,
    NOTE_C5, NOTE_CS5, NOTE_D5, NOTE_DS5, NOTE_E5, NOTE_F5, NOTE_FS5, NOTE_G5, NOTE_GS5, NOTE_A5, NOTE_AS5, NOTE_B5,
    NOTE_C6, NOTE_CS6, NOTE_D6, NOTE_DS6, NOTE_E6, NOTE_F6, NOTE_FS6, NOTE_G6, NOTE_GS6, NOTE_A6, NOTE_AS6, NOTE_B6,
    NOTE_C7, NOTE_CS7, NOTE_D7, NOTE_DS7, NOTE_E7, NOTE_F7, NOTE_FS7, NOTE_G7, NOTE_GS7, NOTE_A7, NOTE_AS7, NOTE_B7
)

// for optional icd debugger _trap() support
#option _RTTTL_TRAP = false
#if (_RTTTL_TRAP)
  #if not defined(MPLAB_ICD) or (MPLAB_ICD = false)
    #error "_RTTTL_TRAP requires '#option MPLAB_ICD = true'"
  #endif
  include "mplabicd.bas"
#endif

//
//-----------------------------------------------------------------
// table access and misc helper functions
//-----------------------------------------------------------------
//
inline function TBLRD_() as TABLAT
    asm-
        TBLRD*
    end asm
end function

inline function TBLRD_POSTINC() as TABLAT
    asm-
        TBLRD*+
    end asm
end function

macro save_TBLPTR(dw)
    dw.bytes(2) = TBLPTRU
    dw.bytes(1) = TBLPTRH
    dw.bytes(0) = TBLPTRL
end macro

macro restore_TBLPTR(dw)
    TBLPTRU = dw.bytes(2)
    TBLPTRH = dw.bytes(1)
    TBLPTRL = dw.bytes(0)
end macro

// string numeric functions

// check for ascii numeric char 0-9
private function isdigit(n as byte) as boolean
    result = false
    if (n >= byte("0")) and (n <= byte("9")) then
        result = true
    endif
end function

// get a digit char from the string and convert it
// advance TBLPTR to next char
private function get_digit() as byte
    result = TBLRD_POSTINC() - byte("0")
end function

// get a number from the string at the current position
// leaves TBLPTR pointing to first char after the digits
private function get_num() as word
    dim num as word
    
    num = 0
    while (isdigit(TBLRD_()))
        num = word(num * 10) + get_digit()
    end while
    result = num
end function

// get current char and convert to lower case
// does not advance TBLPTR
private function get_char() as TABLAT
    TBLRD_()
    if (TABLAT >= byte("A")) and (TABLAT <= byte("Z")) then
        TABLAT.bits(5) = 1      // TABLAT or $20 converts upper -> lower case
    endif        
end function

// incr char pointer
private inline sub incr_ptr()
    TBLRD_POSTINC()
end sub

// skip over space chars
private sub skip_space()
    while (TBLRD_() = byte(" "))
        TBLRD_POSTINC()
    end while
end sub

// check to see if the tone pin is set to output mode
// if not, assume it hasn't been setup and call tone.init
// (TONE_OUTPUT_PIN and _TONE_OUTPUT_TRIS are tone.bas module #options)
private sub check_tone_init()
    dim tone_tris as _TONE_OUTPUT_TRIS.TONE_OUTPUT_PIN@

    if (tone_tris = 1) then     // pin is not an output...
        tone.init()             // so setup tone module
    endif        
end sub

//
//-----------------------------------------------------------------
// get_name()
//-----------------------------------------------------------------
// get song name
// leaves TBLPTR pointing to the defaults section after the name
// spec extension - the spec limits 'name' to 10 chars, but we allow up to 23 (string default)
// supports embedded spaces in the name
public function get_name(const_addr as longword) as string
    dim name as string,
        ix as byte
    
    // load TBLPTR with the 24-bit addr of the const string
    TBLPTRU = const_addr.bytes(2)
    TBLPTRH = const_addr.bytes(1)
    TBLPTRL = const_addr.bytes(0)

    // skip any initial whitespace
    skip_space()

    // copy string up until the first colon
    ix = 0
    FSR0 = addressof(name)
    while (TBLRD_() <> byte(":"))
        if (TABLAT = 0) then        // hit end of input string
            break
        endif
        if (ix = (sizeof(name)-1)) then     // hit max number of chars
            break
        endif
        ix = ix + 1            
        POSTINC0 = TABLAT           // copy char to output string
        TBLRD_POSTINC()
    end while
    POSTINC0 = 0                    // null terminate the string
    TBLRD_POSTINC()                 // skip ':'
    result = name
end function

//
//-----------------------------------------------------------------
// play()
//-----------------------------------------------------------------
//
// play an RTTTL const string
// parameters:
//  const_addr - address of const string in program memory (ie '@song')
public sub play(const_addr as longword)
    dim t_tblptru as byte,      // copy of TBLPTRU
        t_tblptr as longword    // copy of TBLPTR

    dim 
        default_duration as byte,
        default_octave as byte,
        bpm as word,
        wholenote as longword,      // must be large enough for (60 * 1000 * 4)
        duration as word,
        freq as word,
        scale as byte,
        note as byte,
        num as word,
        special_duration as boolean

    // save tblptru
    t_tblptru = TBLPTRU
    
    // check tone module setup
    check_tone_init()

    // minimal/no error checking in here...
    // RTTTL header format: "name:d=N,o=N,b=NNN:"
    // if not specified, these are the RTTTL defaults
    default_duration = 4        // "d="
    default_octave = 6          // "o="
    bpm = 63                    // "b="

    // load TBLPTR with start of const string and scan past the name
    get_name(const_addr)
    
    // scan off any initial whitespace following the header
    skip_space()

    // get default duration
    if (get_char() = byte("d")) then
        incr_ptr()                          // skip "d="
        incr_ptr()
        num = get_num()
        if (num > 0) then
            default_duration = num
        endif            
        incr_ptr()                          // skip comma
    endif
    skip_space()

    // get default octave
    if (get_char() = byte("o")) then
        incr_ptr()                          // skip "o="
        incr_ptr()
        num = get_digit()
        if (num >= 4) and (num <= 7) then
            default_octave = num
        endif            
        incr_ptr()                          // skip comma
    endif
    skip_space()

    // get BPM
    if (get_char() = byte("b")) then
        incr_ptr()                          // skip "b="
        incr_ptr()
        bpm = get_num()
        incr_ptr()                          // skip colon
    endif
    skip_space()

    // BPM usually expresses the number of quarter notes per minute
    wholenote = (60 * 1000 * 4) / bpm       // this is the time for whole note (in milliseconds)
    
    // set tone module to blocking mode so we don't have to call wait() for each note
    tone.setwait()

    // begin note loop
    while (get_char() <> 0)
        skip_space()
        // get note duration (if available)
        num = get_num()
        if (num <> 0) then
            duration = wholenote / num
        else 
            duration = wholenote / default_duration
        endif

        // get the note "a"-"g" or "p" (pause)
        select (get_char())
            case byte("c")
                note = 1
            case byte("d")
                note = 3
            case byte("e")
                note = 5
            case byte("f")
                note = 6
            case byte("g")
                note = 8
            case byte("a")
                note = 10
            case byte("b"), byte("h")       // allow european scale
                note = 12
            case byte("p")                  // pause
                note = 0
            else
                note = 0
        end select
        incr_ptr()                          // point to next char

        // get optional '#' sharp
        if (get_char() = byte("#")) then
            note = note + 1
            incr_ptr()
        endif

        // get optional '-' flat (spec extension)
        if (get_char() = byte("-")) then
            note = note - 1
            incr_ptr()
        endif

        // spec extension - we allow the optional special-duration '.' 
        // specifier either before the scale or after
        special_duration = false
        if (get_char() = byte(".")) then
            special_duration = true
            incr_ptr()
        endif
  
        // get optional scale (aka octave)
        if (isdigit(get_char())) then
            scale = get_digit()
        else
            scale = default_octave
        endif
        scale = scale + OCTAVE_OFFSET

        // get optional special-duration '.' dotted note (this is where it's supposed to be)
        if (get_char() = byte(".")) then
            special_duration = true
            incr_ptr()
        endif
        // a dotted note increases time by 1.5
        if (special_duration) then
            duration = duration + (duration/2)
        endif
  
        // skip comma for next note (or we may be at the end)
        if (get_char() = byte(",")) then
            incr_ptr()
        endif
        skip_space()

        // play the note
        if (note <> 0) then
            save_TBLPTR(t_tblptr)       // save our TBLPTR settings
            scale = (scale - DEFAULT_SCALE) * NOTES_PER_SCALE
            note = note + scale         // add proper scale to the base note setting
            freq = notes(note)          // get note freq from table 
          #if (_RTTTL_TRAP)
            _trap()                     // halt debugger
          #endif
            tone.play(freq, duration)   // now play it
            restore_TBLPTR(t_tblptr)    // restore our TBLPTR settings
          #if (_RTTTL_TRAP)
            _trap()                     // halt debugger
          #endif
        else                            // note = 0... pause x msecs
            delayms(duration)
        endif
    end while

    // restore tblptru
    TBLPTRU = t_tblptru
end sub

end module
